import itertools
import requests
import warnings
import numpy as np

from typing import List

from doc_utils import DocUtils

BASE_2VEC_DEFINITON = {
    "vector_length": None,
    "description": None,
    "paper": None,
    "repo": None,
    "model_name": None,
    "architecture": None,
    "tasks": None,
    "limitations": None,
    "download_required": None,
    "training_required": None,
    "finetunable": None,
}


class Base2Vec(DocUtils):
    """
    Base class for vector
    """

    def __init__(self):
        self.__dict__.update(BASE_2VEC_DEFINITON)

    @classmethod
    def validate_model_url(cls, model_url: str, list_of_urls: List[str]):
        """
        Validate the model url belongs in the list of urls. This is to help
        users to avoid mis-spelling the name of the model.

        # TODO:
        Improve model URL validation to not include final number in URl string.

        Args:
            model_url: The URl of the the model in question
            list_of_urls: The list of URLS for the model in question

        """
        if model_url in list_of_urls:
            return True

        if "tfhub" in model_url:
            # If the url has a number in it then we can take that into account
            for url in list_of_urls:
                if model_url[:-1] in url:
                    return True
        # TODO: Write documentation link to debugging the Model URL.
        warnings.warn(
            "We have not tested this url. Please use URL at your own risk."
            + "Please use the is_url_working method to test if this is a working url if "
            + "this is not a local directory.",
            UserWarning,
        )

    @staticmethod
    def is_url_working(url):
        response = requests.head(url)
        if response.status_code == 200:
            return True
        return False

    @classmethod
    def chunk(self, lst: List, chunksize: int):
        """
        Chunk an iterable object in Python but not a pandas DataFrame.
        Args:
            lst:
                Python List
            chunksize:
                The chunk size of an object.
        Example:
            >>> documents = [{...}]
            >>> ViClient.chunk(documents)
        """
        for i in range(0, len(lst), chunksize):
            yield lst[i : i + chunksize]

    def _vector_operation(self, vectors, vector_operation: str = "mean", axis=0):
        """
        Args:
            Vectors: the list of vectors to include
            vector_operation: One of ['mean', 'minus', 'sum', 'min', 'max']
            axis: The axis to which to perform the operation
        """
        if vector_operation == "mean":
            return np.mean(vectors, axis=axis).tolist()
        elif vector_operation == "minus":
            return np.subtract(vectors, axis=axis).tolist()
        elif vector_operation == "sum":
            return np.sum(vectors, axis=axis).tolist()
        elif vector_operation == "min":
            return np.min(vectors, axis=axis).tolist()
        elif vector_operation == "max":
            return np.max(vectors, axis=axis).tolist()
        else:
            return np.mean(vectors, axis=axis).tolist()

    @property
    def __name__(self):
        """
        Return the name of the model. If name is not set, returns the
        model_id.
        """
        if hasattr(self, "_name"):
            return self._name.replace("-", "_")
        elif hasattr(self, "definition"):
            if "/" in self.definition.model_id:
                return self.definition.model_id.split("/")[1].replace("-", "_")
            return self.definition.model_id
        return ""

    @__name__.setter
    def __name__(self, value):
        """
        Set the name.
        """
        setattr(self, "_name", value)

    @property
    def zero_vector(self):
        if hasattr(self, "vector_length"):
            return self.vector_length * [1e-7]
        else:
            raise ValueError("Please set attribute vector_length")

    def is_empty_vector(self, vector):
        return all([x == 1e-7 for x in vector])

    def get_default_vector_field_name(self, field, field_type="vector"):
        if field_type == "vector":
            return field + "_" + self.__name__ + "_vector_"
        elif field_type == "chunkvector":
            return field + "_" + self.__name__ + "_chunkvector_"

    def _encode_document(
        self,
        field,
        doc,
        vector_error_treatment="zero_vector",
        field_type: str = "vector",
    ):
        """Encode document"""
        vector = self.encode(self.get_field(field, doc))
        if vector_error_treatment == "zero_vector":
            self.set_field(
                self.get_default_vector_field_name(field, field_type=field_type),
                doc,
                vector,
            )
            return
        elif vector_error_treatment == "do_not_include":
            return
        else:
            if vector is None or self.is_empty_vector(vector):
                vector = vector_error_treatment
            self.set_field(self.get_default_vector_field_name(field), doc, vector)

    def _encode_chunk_document(
        self,
        chunk_field,
        field,
        doc,
        vector_error_treatment="zero_vector",
        field_type: str = "chunkvector",
    ):
        """Encode a chunk document"""
        chunk_docs = self.get_field(chunk_field, doc)
        if hasattr(self, "bulk_encode"):
            return self.encode_documents_in_bulk(
                [field],
                chunk_docs,
                field_type=field_type,
                vector_error_treatment=vector_error_treatment,
            )
        elif hasattr(self, "encode"):
            return self.encode_documents(
                [field],
                chunk_docs,
                field_type=field_type,
                vector_error_treatment=vector_error_treatment,
            )

    def _bulk_encode_document(
        self,
        field,
        docs,
        vector_error_treatment: str = "zero_vector",
        field_type="vector",
    ):
        """bulk encode documents"""
        vectors = self.bulk_encode(self.get_field_across_documents(field, docs))
        if vector_error_treatment == "zero_vector":
            self.set_field_across_documents(
                self.get_default_vector_field_name(field, field_type=field_type),
                vectors,
                docs,
            )
            return
        elif vector_error_treatment == "do_not_include":
            [
                self.set_field(
                    self.get_default_vector_field_name(field, field_type=field_type),
                    value=vectors[i],
                    doc=d,
                )
                for i, d in enumerate(docs)
                if not self.is_empty_vector(vectors[i])
            ]
        else:
            [
                self.set_field(
                    self.get_default_vector_field_name(field, field_type=field_type), d
                )
                if not self.is_empty_vector(vectors[i])
                else vector_error_treatment
                for i, d in enumerate(docs)
            ]
            return

    def encode_documents(
        self,
        fields: list,
        documents: list,
        vector_error_treatment="zero_vector",
        field_type="vector",
    ):
        """
        Encode documents and their specific fields. Note that this runs off the
        default `encode` method. If there is a specific function that you want run, ensure
        that it is set to the encode function.

        Parameters:
            missing_treatment:
                Missing treatment can be one of ["do_not_include", "zero_vector", value].
            documents:
                The documents that are being used
            fields:
                The list of fields to be used
            field_type:
                Accepts "vector" or "chunkvector"
        """
        for f in fields:
            # Replace with case-switch in future
            [
                self._encode_document(
                    f,
                    d,
                    vector_error_treatment=vector_error_treatment,
                    field_type=field_type,
                )
                for d in documents
                if self.is_field(f, d)
            ]
        return documents

    def encode_chunk_documents(
        self,
        chunk_field,
        fields: list,
        documents: list,
        vector_error_treatment: str = "zero_vector",
    ):
        """Encode chunk documents. Loops through every field and then every document.

        Parameters:
            chunk_field: The field for chunking
            fields: A list of fields for chunk documents
            documents: a list of documents
            vector_error_treatment: Vector Error Treatment

        Example:
            >>> chunk_docs = enc.encode_chunk_documents(chunk_field="value", fields=["text"], documents=chunk_docs)

        """
        # Replace with case-switch in future
        for f in fields:
            [
                self._encode_chunk_document(
                    chunk_field=chunk_field,
                    field=f,
                    doc=d,
                    vector_error_treatment=vector_error_treatment,
                    field_type="chunkvector",
                )
                for d in documents
                if self.is_field(chunk_field, d)
            ]
        return documents

    def encode_documents_in_bulk(
        self,
        fields: list,
        documents: list,
        vector_error_treatment="zero_vector",
        field_type="vector",
    ):
        """
        Encode documents and their specific fields. Note that this runs off the
        default `encode` method. If there is a specific function that you want run, ensure
        that it is set to the encode function.

        Parameters:
            missing_treatment:
                Missing treatment can be one of ["do_not_include", "zero_vector", value].
            documents:
                The documents that are being used
            fields:
                The list of fields to be used
        """
        for f in fields:
            # Replace with case-switch in future
            contained_docs = [d for d in documents if self.is_field(f, d)]
            self._bulk_encode_document(
                f,
                contained_docs,
                vector_error_treatment=vector_error_treatment,
                field_type=field_type,
            )
        return documents

    def get_combinations(self, lst, maximum_span=5):
        for i, j in itertools.combinations(range(len(lst) + 1), 2):
            if j - -i == len(lst):
                continue
            if j - i <= maximum_span:
                yield lst[i:j]

    def get_cosine_similarity(self, vector_1, vector_2):
        from scipy import spatial

        return 1 - spatial.distance.cosine(vector_1, vector_2)

    def get_word_combinations(self, sentence, maximum_span=5):
        for x in list(self.get_combinations(sentence.split(), maximum_span)):
            yield " ".join(x)
        yield ""

    def get_result(self, result_text, query_vector):
        return {
            "_search_score": self.get_cosine_similarity(
                query_vector, self.encode(result_text)
            ),
            "text": result_text,
        }

    def explain(
        self,
        query_text,
        result_texts,
        highlight_threshold=0.5,
        max_highlights=None,
        return_cos_similarity_docs: bool = True,
        ignore_warnings: bool = True,
    ):
        import numpy as np
        from IPython.display import display
        from sklearn.preprocessing import MinMaxScaler
        from jsonshower import show_json

        query_vector = self.encode(query_text)
        cos_similarity = list()
        for result_text in result_texts:
            doc = self.get_result(result_text, query_vector)
            doc["explain_chunk"] = [
                {"text": t} for t in self.get_word_combinations(result_text)
            ]
            self.encode_documents(["text"], doc["explain_chunk"])
            for d in doc["explain_chunk"]:
                d["_search_score"] = self.get_cosine_similarity(
                    d[f"text_{self.__name__}_vector_"], query_vector
                )
            doc["explain_chunk"] = sorted(
                doc["explain_chunk"], key=lambda x: x["_search_score"], reverse=True
            )
            cos_similarity.append(doc.copy())

        if return_cos_similarity_docs:
            return cos_similarity

        for r in cos_similarity:
            scaler = MinMaxScaler()
            normed_scores = scaler.fit_transform(
                np.array(
                    [np.exp(_r["_search_score"]) for _r in r["explain_chunk"]]
                ).reshape((-1, 1))
            )
            highlight_colors = [f"rgba(253, 253, 150, {x[0]})" for x in normed_scores]
            text_fields = [
                "explain_chunk." + str(i) + ".text"
                for i in range(len(r["explain_chunk"]))
                if normed_scores[i][0] > 0.5
            ]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                if max_highlights is None:
                    display(
                        show_json(
                            [r],
                            highlight_fields={"text": text_fields},
                            highlight_colors=highlight_colors,
                            return_html=False,
                            show_highlight_headers=False,
                            case_insensitive_highlighting=True,
                        )
                    )
                else:
                    display(
                        show_json(
                            [r],
                            highlight_fields={"text": text_fields[:max_highlights]},
                            highlight_colors=highlight_colors,
                            return_html=False,
                            show_highlight_headers=False,
                            case_insensitive_highlighting=True,
                        )
                    )



"""
Extract QA to generate summary or contextual question using the T5 model.
"""
from relevanceai.constants.errors import MissingPackageError

try:
    from transformers import pipeline
except:
    raise MissingPackageError("transformers[sentencepiece]")

from relevanceai.operations.base import BaseOps


class QAOps(BaseOps):
    """
    QAOps
    """

    def __init__(self, model_name: str = "mrm8488/deberta-v3-base-finetuned-squadv2"):
        self.model_name = model_name
        self.qa_model = pipeline("question-answering", model=model_name)

    def question_answer(self, question: str, context: str):
        output = self.qa_model(question=question, context=context)
        return {"answer": output["answer"], "score": output["score"]}

    def bulk_question_answer(self, question, contexts):
        return [
            self.qa_model(question=question, context=context) for context in contexts
        ]



from typing import (
    Any,
    Set,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np

from relevanceai.client.helpers import Credentials
from relevanceai.operations import BaseOps
from relevanceai.utils.decorators import beta, track, deprecated
from relevanceai.constants import (
    Warning,
    Messages,
    CLUSTER_APP_LINK,
    ModelNotSupportedError,
)
from relevanceai.operations.cluster.models.summarizer import TransformersLMSummarizer

from relevanceai.operations.cluster.utils import ClusterUtils
from doc_utils import DocUtils


class ClusterOps(ClusterUtils, BaseOps, DocUtils):
    """
    You can load ClusterOps instances in 2 ways.

    .. code-block::

        # State the vector fields and alias in the ClusterOps object
        cluster_ops = client.ClusterOps(
           alias="kmeans-16",
            dataset_id="sample_dataset_id",
            vector_fields=['sample_vector_']
        )

        cluster_ops.list_closest()

        # State the vector fields and alias in the operational call
        cluster_ops = client.ClusterOps(alias="kmeans-16")
        cluster_ops.list_closest(
            dataset="sample_dataset_id",
            vector_fields=["documentation_vector_]
        )

    """

    dataset_id: str
    cluster_field: str

    def __init__(
        self,
        credentials: Credentials,
        model: Any = None,
        alias: str = None,
        n_clusters: Optional[int] = None,
        cluster_config: Optional[Dict[str, Any]] = None,
        outlier_value: int = -1,
        outlier_label: str = "outlier",
        verbose: bool = True,
        vector_fields: Optional[list] = None,
        **kwargs,
    ):
        """
        ClusterOps object

        Parameters
        -------------

        model: Any
            The string of clustering algorithm, class of clustering algorithm or custom clustering class.
            If custom, the model must contain the method for fit_predict and must output a numpy array for labels

        alias: str = None
            An alias to be given for clustering.
            If no alias is provided, alias becomes the cluster model name hypen n_clusters e.g. (kmeans-5, birch-8)

        n_clusters: Optional[int] = None
            Number of clusters to find, should the argument be required of the algorithm.
            If None, n_clusters defaults to 8 for sklearn models ONLY.

        cluster_config: Optional[Dict[str, Any]] = None
            Hyperparameters for the clustering model.
            See the documentation for supported clustering models for list of hyperparameters

        outlier_value: int = -1
            For unsupervised clustering models like OPTICS etc. outliers are given the integer label -1

        outlier_label: str = "outlier"
            When viewing the cluster app dashboard, outliers will be prefixed with outlier_label

        """

        self.cluster_config = {} if cluster_config is None else cluster_config  # type: ignore

        self.model_name = None

        if model is None:
            model = "community_detection"
            if verbose:
                print(f"No clustering model selected: defaulting to `{model}`")

        self.n_clusters = n_clusters

        if n_clusters is not None:
            self.cluster_config["n_clusters"] = n_clusters  # type: ignore

        self.model = self._get_model(model)

        if self.n_clusters is None:
            if hasattr(self.model, "n_clusters"):
                self.n_clusters = self.model.n_clusters

            elif hasattr(self.model, "k"):
                self.n_clusters = self.model.k

        if "package" not in self.__dict__:
            self.package = self._get_package(self.model)

        self.alias = self._get_alias(alias)
        self.vector_fields = vector_fields  # type: ignore
        if self.vector_fields is not None and len(self.vector_fields) >= 1:
            self.vector_field = self.vector_fields[0]
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label

        for key, value in kwargs.items():
            if not hasattr(self, key):
                if key == "vector_fields":
                    setattr(self, "vector_field", value[0])
                setattr(self, key, value)

        super().__init__(credentials)

    def __call__(
        self,
        dataset_id: str,
        vector_fields: Optional[List[str]] = None,
        include_cluster_report: bool = True,
        **kwargs,
    ) -> None:
        return self.run(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            include_cluster_report=include_cluster_report,
            **kwargs,
        )

    @deprecated(version="1.0.0")
    def fit_predict_update(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def _get_schema(self) -> Dict:
        return self.datasets.schema(dataset_id=self.dataset_id)

    def _get_alias(self, alias: Any) -> str:
        # Auto-generates alias here
        if alias is None:
            if hasattr(self.model, "n_clusters"):
                n_clusters = (
                    self.n_clusters
                    if self.n_clusters is not None
                    else self.model.n_clusters
                )
                alias = f"{self.model_name}-{n_clusters}"

            elif hasattr(self.model, "k"):
                n_clusters = (
                    self.n_clusters if self.n_clusters is not None else self.model.k
                )
                alias = f"{self.model_name}-{n_clusters}"

            else:
                alias = self.model_name

            Warning.MISSING_ALIAS.format(alias=alias)

        print(f"The alias is `{alias.lower()}`.")
        return alias.lower()

    def _get_package(self, model):
        model_name = str(model.__class__).lower()
        if "function" in model_name:
            model_name = str(model.__name__)

        if "sklearn" in model_name:
            package = "sklearn"

        elif "faiss" in model_name:
            package = "faiss"

        elif "hdbscan" in model_name:
            package = "hdbscan"

        elif "communitydetection" in model_name:
            package = "sentence-transformers"

        else:
            package = "custom"

        return package

    def _get_model(self, model):
        if isinstance(model, str):
            model = model.lower().replace(" ", "").replace("_", "")
            self.model_name = model
        else:
            self.model_name = str(model.__class__).split(".")[-1].split("'>")[0]
            self.package = "custom"
            return model

        if isinstance(model, str):
            if model == "affinitypropagation":
                from sklearn.cluster import AffinityPropagation

                model = AffinityPropagation(**self.cluster_config)

            elif model == "agglomerativeclustering":
                from sklearn.cluster import AgglomerativeClustering

                model = AgglomerativeClustering(**self.cluster_config)

            elif model == "birch":
                from sklearn.cluster import Birch

                model = Birch(**self.cluster_config)

            elif model == "dbscan":
                from sklearn.cluster import DBSCAN

                model = DBSCAN(**self.cluster_config)

            elif model == "optics":
                from sklearn.cluster import OPTICS

                model = OPTICS(**self.cluster_config)

            elif model == "kmeans":
                from sklearn.cluster import KMeans

                model = KMeans(**self.cluster_config)

            elif model == "featureagglomeration":
                from sklearn.cluster import FeatureAgglomeration

                model = FeatureAgglomeration(**self.cluster_config)

            elif model == "meanshift":
                from sklearn.cluster import MeanShift

                model = MeanShift(**self.cluster_config)

            elif model == "minibatchkmeans":
                from sklearn.cluster import MiniBatchKMeans

                model = MiniBatchKMeans(**self.cluster_config)

            elif model == "spectralclustering":
                from sklearn.cluster import SpectralClustering

                model = SpectralClustering(**self.cluster_config)

            elif model == "spectralbiclustering":
                from sklearn.cluster import SpectralBiclustering

                model = SpectralBiclustering(**self.cluster_config)

            elif model == "spectralcoclustering":
                from sklearn.cluster import SpectralCoclustering

                model = SpectralCoclustering(**self.cluster_config)

            elif model == "hdbscan":
                from hdbscan import HDBSCAN

                model = HDBSCAN(**self.cluster_config)

            elif model in "communitydetection":
                from relevanceai.operations.cluster.algorithms import CommunityDetection

                model = CommunityDetection(**self.cluster_config)

            elif "faiss" in model:
                from faiss import Kmeans

                model = Kmeans(**self.cluster_config)
            else:
                raise ValueError(
                    """Invalid model. This should be one of ['affinitypropagation',
                    'agglomerativeclustering', 'birch', dbscan', 'optics', 'kmeans',
                    'featureagglomeration', 'meanshift', 'minibatchkmeans',
                    'spectralclustering', 'spectralbiclustering', 'spectralcoclustering',
                    'hdbscan', 'communitydetection']
                    ]"""
                )

        else:
            raise ModelNotSupportedError

        return model

    def _format_labels(self, labels: np.ndarray) -> List[str]:
        labels = labels.flatten().tolist()
        cluster_labels = [
            f"cluster-{str(label)}"
            if label != self.outlier_value
            else self.outlier_label
            for label in labels
        ]
        return cluster_labels

    def _get_centroid_documents(
        self, vectors: np.ndarray, labels: List[str], vector_field: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        centroid_documents = []

        centroids: Dict[str, Any] = {}
        for label, vector in zip(labels, vectors.tolist()):
            if label not in centroids:
                centroids[label] = []
            centroids[label].append(vector)

        for centroid, vectors in centroids.items():
            centroid_vector = np.array(vectors).mean(0).tolist()
            if vector_field:
                centroid_document = {
                    "_id": centroid,
                    vector_field: centroid_vector,
                }
            else:
                centroid_document = dict(
                    _id=centroid,
                    centroid_vector=centroid_vector,
                )
            centroid_documents.append(centroid_document)

        return centroid_documents

    def _get_document_vector_field(self):
        vector_fields = []
        metadata = self.datasets.metadata(self.dataset_id)
        most_recent_updated_vector = max(
            metadata["results"]["_vector_"].items(), key=lambda x: x[1]
        )[0]
        vector_fields.append(most_recent_updated_vector)
        return vector_fields

    def _insert_centroids(
        self,
        dataset_id: str,
        vector_fields: List[str],
        centroid_documents: List[Dict[str, Any]],
    ) -> None:
        self.services.cluster.centroids.insert(
            dataset_id=dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=vector_fields,
            alias=self.alias,
        )

    def _fit_predict(
        self, documents: List[Dict[str, Any]], vector_field: str
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        vectors = np.array(
            [self.get_field(vector_field, document) for document in documents]
        )

        if self.package == "sklearn":
            labels = self.model.fit_predict(vectors)

        elif self.package == "hdbscan":
            self.model.fit(vectors)
            labels = self.model.labels_.reshape(-1, 1)

        elif self.package == "faiss":
            vectors = vectors.astype("float32")
            self.model.train(vectors)
            labels = self.model.assign(vectors)[1]

        elif self.package == "sentence-transformers":
            # TODO: make teh config here better
            labels = self.model(vectors)
            labels = np.array(labels)

        elif self.package == "custom":
            labels = self.model.fit_predict(vectors)
            if not isinstance(labels, np.ndarray):
                raise ValueError("Custom model must output np.ndarray for labels")

            if labels.shape[0] != vectors.shape[0]:
                raise ValueError("incorrect shape")

        labels = self._format_labels(labels)

        if self.n_clusters is None:
            n_clusters = len(set(labels)) - 1
            print(f"Found {n_clusters} clusters using {self.model_name}")

        self.set_field_across_documents(
            field=self.cluster_field,
            values=labels,
            docs=documents,
        )

        centroid_documents = self._get_centroid_documents(
            vectors, labels, vector_field=vector_field
        )

        return centroid_documents, documents

    def _print_app_link(self):
        link = CLUSTER_APP_LINK.format(self.dataset_id)
        print(Messages.BUILD_HERE + link)

    @track
    def run(
        self,
        dataset_id: str,
        vector_fields: Optional[List[str]] = None,
        filters: Optional[list] = None,
        show_progress_bar: bool = True,
        verbose: bool = True,
        include_cluster_report: bool = True,
        report_name: str = "cluster-report",
    ) -> None:
        """
        Run clustering on a dataset

        Parameters
        --------------

        dataset_id: Optional[Union[str, Any]]
            The dataset ID
        vector_fields: Optional[List[str]]
            List of vector fields
        show_progress_bar: bool
            If True, the progress bar can be shown

        """
        filters = [] if filters is None else filters
        if not isinstance(dataset_id, str):
            if hasattr(dataset_id, "dataset_id"):
                dataset_id = dataset_id.dataset_id  # type: ignore
        self.dataset_id = dataset_id

        if vector_fields is None:
            vector_fields = self._get_document_vector_field()
            print(f"No vector_field given: defaulting to {vector_fields}")

        vector_field = vector_fields[0]
        self.cluster_field = f"_cluster_.{vector_fields[0]}.{self.alias}"

        # get all documents
        print("Retrieving all documents...")
        from relevanceai.utils.filter_helper import create_filter

        filters = create_filter(vector_field, filter_type="exists")
        documents = self._get_all_documents(
            dataset_id=dataset_id,
            select_fields=vector_fields,
            show_progress_bar=show_progress_bar,
            include_vector=True,
            filters=filters,
        )

        # fit model, predict and label all documents
        print("Predicting on all documents...")
        centroid_documents, labelled_documents = self._fit_predict(
            documents=documents,
            vector_field=vector_field,
        )

        print("Updating cluster labels...")
        results = self._update_documents(
            dataset_id=dataset_id,
            documents=labelled_documents,
            show_progress_bar=show_progress_bar,
        )

        print("Inserting Centroids...")
        self._insert_centroids(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            centroid_documents=centroid_documents,
        )
        if include_cluster_report:
            print("Generating evaluation report for your clusters…")
            from relevanceai.reports.cluster.report import ClusterReport

            centroids = self.get_field_across_documents(
                vector_field, centroid_documents, missing_treatment="raise_error"
            )

            X = self.get_field_across_documents(
                vector_field, documents, missing_treatment=self.outlier_value
            )

            cluster_labels = self.get_field_across_documents(
                self.cluster_field, documents, missing_treatment=-1
            )

            if len(cluster_labels) != len(X):
                raise ValueError("Damn son. How you like them apples.")

            self.report = ClusterReport(
                X=X,
                cluster_labels=cluster_labels,
                model=self.model,
                outlier_label=-1,
                centroids=centroids,
                verbose=True,
                include_typecheck=False,
            )

            try:
                response = self.reports.clusters.create(
                    name=report_name,
                    report=self.json_encoder(self.report.internal_report),
                )

                if verbose:
                    print(
                        f"📊 You can now access your report at https://cloud.relevance.ai/report/cluster/{self.region}/{response['_id']}"
                    )
            except Exception as e:
                print("Error creating cluster report! " + str(e))

        # link back to dashboard
        if verbose:
            self._print_app_link()

    def cluster_report(self, X, cluster_labels, centroids):
        from relevanceai.reports.cluster.report import ClusterReport

        cluster_labels = self.fit_predict(X)
        centroids = ClusterReport.calculate_centroids(X, cluster_labels)
        report = ...
        response = self.store_cluster_report(
            report_name="kmeans", report=report.internal_report
        )

    @track
    def closest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        verbose: bool = True,
    ):
        """
        List of documents closest from the center.

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_closest_to_center(
            dataset_id=dataset_id,
            vector_fields=[vector_field],
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
            verbose=verbose,
        )

    @staticmethod
    def get_cluster_summary(
        summarizer,
        docs: Dict,
        summarize_fields: List[str],
        max_length: int = 100,
        first_sentence_only: bool = True,
    ):
        def _clean_sentence(s):
            s = (
                s.replace(". .", ".")
                .replace(" .", ".")
                .replace("\n", "")
                .replace("..", ".")
                .strip()
            )
            if s[-1] != ".":
                s += "."
            return s

        cluster_summary = {}
        for cluster, results in docs["results"].items():
            for f in summarize_fields:
                summary_fields = [
                    _clean_sentence(d[f])
                    for d in results["results"]
                    if d.get(f) and d[f] not in [" ", "."]
                ]
                summary_output = summarizer(
                    " ".join(summary_fields), max_length=max_length
                )[0]["summary_text"]
                summary = summary_output.replace(" .", ".").strip()
            if first_sentence_only:
                cluster_summary[cluster] = summary.split(".")[0]
            else:
                cluster_summary[cluster] = summary
        return cluster_summary

    @beta
    @track
    def summarize_closest(
        self,
        summarize_fields: List[str],
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        model_name: str = "philschmid/bart-large-cnn-samsum",
        tokenizer: Optional[str] = None,
        max_length: int = 100,
        deployable_id: Optional[str] = None,
        first_sentence_only: bool = True,
        **kwargs,
    ):
        """
        List of documents closest from the center.

        Parameters
        ----------
        summarize_fields: list
            Fields to perform summarization, empty array/list means all fields
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        model_name: str
            Huggingface Model to use for summarization.
            Pick from
            https://huggingface.co/models?pipeline_tag=summarization&sort=downloadshttps://huggingface.co/models?pipeline_tag=summarization
        tokenizer: str
            Tokenizer to use for summarization, allows you to bring your own tokenizer,
            else will instantiate pre-trained from selected model

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        if not tokenizer:
            tokenizer = model_name

        if not hasattr(self, "summarizer"):
            self.summarizer = TransformersLMSummarizer(model_name, tokenizer, **kwargs)

        center_docs = self.list_closest(
            select_fields=summarize_fields,
            dataset_id=dataset_id,
            vector_field=vector_field,
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

        cluster_summary = self.get_cluster_summary(
            self.summarizer,
            docs=center_docs,
            summarize_fields=summarize_fields,
            max_length=max_length,
            first_sentence_only=first_sentence_only,
        )

        if deployable_id is not None:
            if dataset_id is None:
                if not hasattr(self, "dataset_id"):
                    raise ValueError("You need a dataset ID to update.")
                else:
                    dataset_id = self.dataset_id
            configuration = self.deployables.get(deployable_id=deployable_id)
            configuration["cluster-labels"] = cluster_summary
            self.deployables.update(
                deployable_id=deployable_id,
                dataset_id=dataset_id,
                configuration=configuration,
            )
        return {"results": cluster_summary}

    @track
    def furthest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
    ):
        """
        List documents furthest from the center.

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_furthest_from_center(
            dataset_id=dataset_id,
            vector_fields=[vector_field],
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

    @beta
    @track
    def summarize_furthest(
        self,
        summarize_fields: List[str],
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        model_name: str = "sshleifer/distilbart-cnn-6-6",
        tokenizer: Optional[str] = None,
        **kwargs,
    ):
        """
        List of documents furthest from the center.

        Parameters
        ----------
        summarize_fields: list
            Fields to perform summarization, empty array/list means all fields
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        model_name: str
            Huggingface Model to use for summarization.
            Pick from
            https://huggingface.co/models?pipeline_tag=summarization&sort=downloadshttps://huggingface.co/models?pipeline_tag=summarization
        tokenizer: str
            Tokenizer to use for summarization, allows you to bring your own tokenizer,
            else will instantiate pre-trained from selected model

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        if not tokenizer:
            tokenizer = model_name
        summarizer = TransformersLMSummarizer(model_name, tokenizer)

        center_docs = self.list_furthest(
            select_fields=summarize_fields,
            dataset_id=dataset_id,
            vector_field=vector_field,
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

        cluster_summary = self.get_cluster_summary(
            summarizer, docs=center_docs, summarize_fields=summarize_fields
        )

        return {"results": cluster_summary}

    # Convenience functions
    list_closest = closest
    list_furthest = furthest

    def _retrieve_dataset_id(self, dataset) -> str:
        """Helper method to get multiple dataset values.
        Dataset can be either a Dataset or a string object."""
        from relevanceai.dataset import Dataset

        if isinstance(dataset, Dataset):
            dataset_id: str = dataset.dataset_id
        elif isinstance(dataset, str):
            dataset_id = dataset
        elif dataset is None:
            if hasattr(self, "dataset_id"):
                # let's not surprise users
                print(
                    f"No dataset supplied - using last stored one '{self.dataset_id}'."
                )
                dataset_id = str(self.dataset_id)
            else:
                raise ValueError("Please supply dataset.")
        return dataset_id

    @track
    def aggregate(
        self,
        vector_fields: List[str] = None,
        metrics: Optional[list] = None,
        sort: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        dataset=None,
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n
        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }

        For example, one can use the following aggregations to group score based on region and player name.

        .. code-block::

            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }
        "groupby" is the fields you want to split the data into. These are the available groupby types:
            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric
        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:
            - "avg", "max", "min", "sum", "cardinality"
        The response returned has the following in descending order. \n
        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:
            .. code-block::
                {
                    'groupby':[
                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
                        'group_size': 10, 'select_fields': ["name"]},
                    ],
                    'metrics':[
                        {'name':'Price Average','field':'price','agg':'avg'},
                    ],
                }
                # ouptut example:
                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}
        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten

        Example
        ---------
        .. code-block::
            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.run(df, vector_fields=["sample_vector_"])
            clusterer.aggregate(
                "sample_dataset_id",
                groupby=[{
                    "field": "title",
                    "agg": "wordcloud",
                }],
                vector_fields=['sample_vector_']
            )
        """
        metrics = [] if metrics is None else metrics
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort

        return self.services.cluster.aggregate(
            dataset_id=self._retrieve_dataset_id(dataset),
            vector_fields=self.vector_fields if not vector_fields else vector_fields,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            filters=filters,
            alias=self.alias,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
        )

    @track
    def merge(
        self,
        cluster_labels: List,
        alias: Optional[str] = None,
        show_progress_bar: bool = True,
        **update_kwargs,
    ):
        """
        Parameters
        ----------
        cluster_labels : Tuple[int]
            a tuple of integers representing the cluster ids you would like to merge

        alias: str
            the alias of the clustering you like to merge labels within

        show_progress_bar: bool
            whether or not to show the progress bar

        Example
        -------

        .. code-block::

            dataset.cluster(
                model="kmeans",
                n_clusters=3,
                vector_fields=["sample_1_vector_"],
            )

            ops = ClusterOps.from_dataset(
                dataset=dataset,
                alias="kmeans-3",
                vector_fields=["sample_1_vector_"],
            )

            ops.merge(cluster_labels=(0, 1), alias="kmeans-3")

        """

        if alias is None:
            if hasattr(self, "alias"):
                alias = self.alias
            else:
                raise ValueError("Please specify alias= as it was not detected")

        try:
            centroid_documents = self.services.cluster.centroids.list(
                dataset_id=self.dataset_id,
                vector_fields=[self.vector_field],
                alias=alias,
                include_vector=True,
            )["results"]

            relevant_centroids = [
                centroid[self.vector_field]
                for centroid in centroid_documents
                if any(f"-{cluster}" in centroid["_id"] for cluster in cluster_labels)
            ]
            new_centroid = np.array(relevant_centroids).mean(0).tolist()
            if isinstance(cluster_labels[0], int):
                new_centroid_doc = {
                    "_id": f"cluster-{cluster_labels[0]}",
                    self.vector_field: new_centroid,
                }
            elif isinstance(cluster_labels[0], str):
                if isinstance(cluster_labels[0], int):
                    new_centroid_doc = {
                        "_id": cluster_labels,
                        self.vector_field: new_centroid,
                    }
        except Exception as e:
            print(e)
            pass

        update: dict = {}

        if isinstance(cluster_labels[0], str):
            self.clusters = [cluster for cluster in sorted(cluster_labels)]
            self.min_cluster = cluster_labels[0]
        else:
            self.clusters = [f"cluster-{cluster}" for cluster in sorted(cluster_labels)]
            self.min_cluster = f"cluster-{min(cluster_labels)}"

        print(f"Merging clusters to {cluster_labels[0]}")
        update = {f"_cluster_.{self.vector_field}.{self.alias}": cluster_labels[0]}

        results = self.datasets.documents.update_where(
            dataset_id=self.dataset_id,
            update=update,
            filters=[
                {
                    "field": self._get_cluster_field_name(alias=alias),
                    "filter_type": "categories",
                    "condition": "==",
                    "condition_value": cluster_labels[1:],
                }
            ],
        )
        print(results)

        try:
            # If there are no centroids - move on
            self.services.cluster.centroids.update(
                dataset_id=self.dataset_id,
                vector_fields=[self.vector_field],
                alias=alias,
                cluster_centers=[new_centroid_doc],
            )

            cluster: int

            for cluster in cluster_labels[1:]:
                if isinstance(cluster, str):
                    centroid_id = cluster
                else:
                    centroid_id = f"cluster-{cluster}"
                self.services.cluster.centroids.delete(
                    dataset_id=self.dataset_id,
                    centroid_id=centroid_id,
                    alias=self.alias,
                    vector_fields=[self.vector_field],
                )
        except:
            pass

    def create_centroids(self):
        """
        Calculate centroids

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                alias="kmeans-40",
                vector_fields=['text_roberta-large_vector_']
            )
            centroids = cluster_ops.create_centroids()

        """
        # Get an array of the different vectors
        if len(self.vector_fields) > 1:
            raise NotImplementedError(
                "Do not currently support multiple vector fields for centroid creation."
            )
        centroid_vectors = {}

        def calculate_centroid(vectors):
            X = np.array(vectors)
            return X.mean(axis=0)

        centroid_vectors = self._operate_across_clusters(
            field=self.vector_fields[0], func=calculate_centroid
        )

        # Does this insert properly?
        if isinstance(centroid_vectors, dict):
            centroid_vectors = [
                {"_id": k, self.vector_fields[0]: v}
                for k, v in centroid_vectors.items()
            ]
        self._insert_centroids(
            dataset_id=self.dataset_id,
            vector_fields=[self.vector_fields[0]],
            centroid_documents=centroid_vectors,
        )
        return centroid_vectors

    @property
    def centroids(self):
        """
        Access the centroids of your dataset easily

        .. code-block::

            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                vector_fields=["sample_vector_"],
                alias="simple"
            )
            cluster_ops.centroids

        """
        if not hasattr(self, "_centroids"):
            self._centroids = self.services.centroids.list(
                dataset_id=self.dataset_id,
                vector_fields=self.vector_fields,
                alias=self.alias,
                page_size=9999,
                include_vector=True,
            )
        return self._centroids

    def insert_centroids(self, centroid_documents):
        """
        Insert your centroids

        Example
        ----------

        .. code-block::

            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                vector_fields=["sample_vector_"],
                alias="simple"
            )
            cluster_ops.insert_centroids(
                [
                    {
                        "_id": "cluster-1",
                        "sample_vector_": [1, 1, 1]
                    }
                ]
            )

        """
        results = self.services.cluster.centroids.insert(
            dataset_id=self.dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )
        return results



from relevanceai.utils.logger import LoguruLogger
from doc_utils import DocUtils
from typing import Dict, List

from relevanceai.constants.errors import MissingPackageError


class TransformersLMSummarizer(LoguruLogger, DocUtils):
    """Seq2seq models using Summarizer class"""

    def __init__(self, model: str, tokenizer: str, *args, **kwargs):
        try:
            from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
        except ModuleNotFoundError as e:
            raise MissingPackageError("transformers")
        try:
            import torch

            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        except ModuleNotFoundError as e:
            raise MissingPackageError("torch")

        if not any([f in model for f in ["t5", "bart"]]):
            raise ValueError(
                "Model must be of t5 or  bart base.\n \
                The models that this pipeline can use are models that have been fine-tuned on a summarization task, \
                which is currently, '`bart-large-cnn`', '`t5-small`', '`t5-base`', '`t5-large`', '`t5-3b`', '`t5-11b`'. \
                See the up-to-date list of available models on \
                    `huggingface.co/models <https://huggingface.co/models?filter=summarization>`__."
            )

        self.model = AutoModelForSeq2SeqLM.from_pretrained(model)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)

    def __call__(self, text: str, max_length: int = 100, num_beams: int = 4):
        self.model = self.model.to(self.device)

        # tokenize without truncation
        inputs_no_trunc = self.tokenizer(
            text, max_length=None, return_tensors="pt", truncation=False
        ).to(self.device)

        # get batches of tokens corresponding to the exact model_max_length
        chunk_start = 0
        chunk_end = self.tokenizer.model_max_length
        inputs_batch_l = []
        while chunk_start <= len(inputs_no_trunc["input_ids"][0]):
            inputs_batch = inputs_no_trunc["input_ids"][0][
                chunk_start:chunk_end
            ]  # get batch of n tokens
            inputs_batch = inputs_batch.reshape(
                (1,) + inputs_batch.shape
            )  ## add batch dimension
            inputs_batch_l.append(inputs_batch)
            chunk_start += self.tokenizer.model_max_length
            chunk_end += self.tokenizer.model_max_length

        # generate a summary on each batch
        summary_ids_l = [
            self.model.generate(
                inputs.to(self.device),
                num_beams=num_beams,
                max_length=max_length,
                early_stopping=True,
            ).to(self.device)
            for inputs in inputs_batch_l
        ]

        # decode the output and join into one string with one paragraph per summary batch
        summary_batch_l = []
        for summary_id in summary_ids_l:
            summary_batch = [
                self.tokenizer.decode(
                    g, skip_special_tokens=True, clean_up_tokenization_spaces=False
                )
                for g in summary_id
            ]
            summary_batch_l.append(summary_batch[0])

        return [{"summary_text": " ".join(summary_batch_l).replace(" .", ".").strip()}]



import pytest


@pytest.mark.skip(reason="Run this to test")
def test_new_config(test_client):
    from relevanceai.utils.datasets import get_ecommerce_dataset_clean
    from relevanceai import Client

    datasets = test_client.list_datasets()
    print(datasets["datasets"][0:2])

    dataset = get_ecommerce_dataset_clean()

    print(len(dataset))
    dataset[0].keys()
    datasets = test_client.list_datasets()
    assert True



from dataclasses import dataclass


def create_filter(
    field: str, filter_type: str, condition: str = "==", condition_value: str = " "
):
    return [
        {
            "field": field,
            "filter_type": filter_type,
            "condition": condition,
            "condition_value": condition_value,
        }
    ]


@dataclass
class Filter:
    field: str
    filter_type: str
    condition: str = "=="
    condition_value: str = " "



import os
import sys


def block_print():
    sys.stdout = open(os.devnull, "w")


def enable_print():
    sys.stdout = sys.__stdout__


def silence(func):
    def wrapper(*args, **kwargs):
        block_print()
        func(*args, **kwargs)
        enable_print()

    return wrapper



import functools
import traceback
import warnings


def catch_errors(func):
    """
    Decorate function and avoid vector errors.
    Example:
        class A:
            @catch_vector_errors
            def encode(self):
                return [1, 2, 3]
    """

    @functools.wraps(func)
    def catch_vector(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            # Bulk encode the functions as opposed to encode to accelerate the
            # actual function call
            if hasattr(func, "__name__"):
                if "bulk_encode" in func.__name__:
                    # Rerun with manual encoding
                    try:
                        encode_fn = getattr(
                            args[0], func.__name__.replace("bulk_encode", "encode")
                        )
                        if len(args) > 1 and isinstance(args[1], list):
                            return [encode_fn(x, **kwargs) for x in args[1]]
                        if kwargs:
                            # Take the first input!
                            for v in kwargs.values():
                                if isinstance(v, list):
                                    return [encode_fn(x, **kwargs) for x in v]
                    except:
                        traceback.print_exc()
                        pass
            warnings.warn("Unable to encode. Filling in with dummy vector.")
            traceback.print_exc()
            # get the vector length from the self body
            vector_length = args[0].vector_length
            if isinstance(args[1], str):
                return [1e-7] * vector_length
            elif isinstance(args[1], list):
                # Return the list of vectors
                return [[1e-7] * vector_length] * len(args[1])
            else:
                return [1e-7] * vector_length
        return

    return catch_vector



"""
Credential-related helpers
"""
from relevanceai.client.helpers import region_to_url


class CredentialsMixin:
    def _set_variables(self):
        (
            self.project,
            self.api_key,
            self.region,
            self.firebase_uid,
        ) = self.credentials.split_token()

        self.base_url = region_to_url(self.region)
        self.base_ingest_url = region_to_url(self.region)

        try:
            self._set_mixpanel_write_key()
        except Exception as e:
            pass



from typing import Dict, Union, Optional, List, Any

from datetime import datetime

import numpy as np
import warnings

from relevanceai._api import APIClient
from relevanceai.client.helpers import Credentials

from relevanceai.constants import Messages
from relevanceai.constants import IMG_EXTS

from relevanceai.utils.decorators import log

from relevanceai.operations.vector.base import Base2Vec


class VectorizeHelpers(APIClient):
    def __init__(self, log_file, credentials: Credentials):
        self.log_file = log_file
        super().__init__(credentials)

    def _get_encoder(self, model: Any) -> Any:
        @log(fn=self.log_file)
        def get_encoder(model):
            if isinstance(model, str):
                model = model.lower().replace(" ", "").replace("_", "")
                model_name = model
            else:
                model_name = str(model.__class__).split(".")[-1].split("'>")[0]
                return model, model_name

            if isinstance(model, str):
                if model == "use":
                    from vectorhub.encoders.text.tfhub import USE2Vec

                    model = USE2Vec()

                elif model == "bert":
                    from vectorhub.encoders.text.tfhub import Bert2Vec

                    model = Bert2Vec()

                elif model == "labse":
                    from vectorhub.encoders.text.tfhub import LaBSE2Vec

                    model = LaBSE2Vec()

                elif model == "elmo":
                    from vectorhub.encoders.text.tfhub import Elmo2Vec

                    model = Elmo2Vec()

                elif model == "clip-image":
                    from vectorhub.bi_encoders.text_image.torch.clip import Clip2Vec

                    model = Clip2Vec()
                    model.encode = model.encode_image

                elif model == "resnet":
                    from vectorhub.encoders.image.fastai import FastAIResnet2Vec

                    model = FastAIResnet2Vec()

                elif model == "mobilenet":
                    from vectorhub.encoders.image.tfhub import MobileNetV12Vec

                    model = MobileNetV12Vec()

                elif model == "clip-text":
                    from vectorhub.bi_encoders.text_image.torch import Clip2Vec

                    model = Clip2Vec()
                    model.encode = model.encode_text

                elif model == "mpnet":
                    from vectorhub.encoders.text.sentence_transformers import (
                        SentenceTransformer2Vec,
                    )

                    model = SentenceTransformer2Vec("all-mpnet-base-v2")
                    model.vector_length = 768

                elif model == "multiqampnet":
                    from vectorhub.encoders.text.sentence_transformers import (
                        SentenceTransformer2Vec,
                    )

                    model = SentenceTransformer2Vec("multi-qa-mpnet-base-dot-v1")
                    model.vector_length = 768

                elif model == "bit":
                    from vectorhub.encoders.image.tfhub import BitMedium2Vec

                    model = BitMedium2Vec()

                else:

                    from vectorhub.encoders.text.sentence_transformers import (
                        SentenceTransformer2Vec,
                    )

                    warnings.warn("Assuming this is a Sentence Transformer model.")
                    model = SentenceTransformer2Vec(model)

            else:
                # TODO: this needs to be referenced from relevance.constants.errors
                raise ValueError("ModelNotSupported")

            assert hasattr(model, "vector_length")
            assert model.vector_length is not None

            model.__name__ = model_name

            return model, model_name

        model, model_name = get_encoder(model)
        self.model_names.append(model_name.lower())
        return model

    def _get_dtype(self, value: Any) -> Union[str, None]:
        if isinstance(value, str):
            if "https://" in value:
                if any(img_ext in value for img_ext in IMG_EXTS):
                    return "_image_"

            elif "$" in value:
                try:
                    float(value.replace("$", ""))
                    return "numeric"

                except:
                    return "_text_"

            elif len(value.strip().split(" ")) > 1:
                return "_text_"

        return None

    def _reduce(self, vectors: np.ndarray, n_components: int = 512) -> np.ndarray:
        from sklearn.decomposition import PCA

        reducer = PCA(
            n_components=min(vectors.shape[0], vectors.shape[1], n_components)
        )
        reduced = reducer.fit_transform(vectors)
        return reduced

    def _get_model_names(self, dtype: str) -> List[Any]:
        if dtype == "_text_":
            if "text" in self.encoders:
                return self.encoders["text"]

            else:
                self.encoders["text"] = ["use"]
                return self.encoders["text"]

        elif dtype == "_image_":
            if "image" in self.encoders:
                return self.encoders["image"]

            else:
                self.encoders["image"] = ["clip"]
                return self.encoders["image"]

        else:
            raise ValueError

    def _get_remaining(self, filters: List[Dict[str, Any]]) -> int:
        return self.get_number_of_documents(dataset_id=self.dataset_id, filters=filters)


class VectorizeOps(VectorizeHelpers):
    def __init__(
        self,
        credentials: Credentials,
        encoders: Optional[Dict[str, List[Any]]] = None,
        log_file: str = "vectorize.logs",
        create_feature_vector: bool = False,
    ):
        super().__init__(log_file=log_file, credentials=credentials)

        self.feature_vector = create_feature_vector
        self.encoders = encoders if encoders is not None else {}
        self.model_names: List[str] = []

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def _get_schema(self):
        return self.datasets.schema(self.dataset_id)

    def _get_detailed_schema(self):
        schema = self.datasets.schema(dataset_id=self.dataset_id)

        text_fields = [column for column, dtype in schema.items() if dtype == "text"]

        for column in text_fields:
            values = self._get_documents(
                dataset_id=self.dataset_id,
                select_fields=[column],
                filters=[
                    {
                        "field": column,
                        "filter_type": "exists",
                        "condition": "==",
                        "condition_value": "",
                    }
                ],
                number_of_documents=1,
            )
            try:
                value = values[0][column]
            except:
                continue

            dtype = self._get_dtype(value)
            if dtype is not None:
                schema[column] = dtype

        return schema

    def _get_numeric_fields(self) -> List[str]:
        numeric_fields = [
            field for field, dtype in self.schema.items() if dtype == "numeric"
        ]
        # facets = self.datasets.facets(self.dataset_id)["results"]
        # stats = {
        #     field: (facets[field]["avg"] - facets[field]["min"])
        #     / (facets[field]["max"] - facets[field]["min"])
        #     for field in [field.replace(" ", "%20") for field in numeric_fields]
        # }
        return numeric_fields

    def _get_fields(self, fields: List[str]) -> Dict[str, str]:
        return {
            field: self.detailed_schema[field]
            for field in fields
            if any(
                dtype in self.detailed_schema[field] for dtype in ["_image_", "_text_"]
            )
        }

    def _validate_fields(self, fields: List[str]):
        foreign_fields = []
        for field in fields:
            if field not in self.schema:
                foreign_fields.append(field)
        else:
            if foreign_fields:
                raise ValueError(
                    f"The following fields are invalid: {', '.join(foreign_fields)}"
                )

    def _init_encoders(self, fields: List[str]):
        dtypes = [self.detailed_schema[field] for field in fields]

        for dtype in ["image", "text"]:
            if f"_{dtype}_" in dtypes:
                models = self._get_model_names(dtype=f"_{dtype}_")
                for index, model in enumerate(models):
                    self.encoders[dtype][index] = self._get_encoder(model=model)

    def _get_vector_fields(self, fields: List[str]) -> List[str]:
        vector_fields = []

        for field in fields:
            dtype = self.detailed_schema[field].replace("_", "")
            encoders: List[Base2Vec] = self.encoders[dtype]
            for encoder in encoders:
                vector_field = encoder.get_default_vector_field_name(field)
                vector_fields.append(vector_field)

        return vector_fields

    def _get_filters(
        self, fields: List[str], vector_fields: List[str]
    ) -> List[Dict[str, Any]]:

        """
        Creates the filters necessary to search all documents
        within a dataset that contain fields specified in "fields"
        but do not contain their resepctive vector_fields defined in "vector_fields"

        e.g.
        fields = ["text", "title"]
        vector_fields = ["text_use_vector_", "title_use_vector_"]

        we want to search the dataset where:
        ("text" * ! "text_use_vector_") + ("title" * ! "title_use_vector_")

        Since the current implementation of filtering only accounts for CNF and not DNF boolean logic,
        We must use boolean algebra here to obtain the CNF from a DNF expression.

        CNF = Conjunctive Normal Form (Sum of Products)
        DNF = Disjunctive Normal Form (Product of Sums)

        This means converting the above to:
        ("text" + "title") * ("text" + ! "title_use_vector_") *
        (! "text_use_vector_" + "title") * (! "text_use_vector_" + ! "title_use_vector_")

        Arguments:
            fields: List[str]
                A list of fields within the dataset

            vector_fields: List[str]
                A list of vector_fields, created from the fields given the current encoders.
                These would be present if the fields in "fields" were vectorized

        Returns:
            filters: List[Dict[str, Any]]
                A list of filters.
        """

        filters = []
        if len(fields) > 1:
            iters = len(fields) ** 2

            for i in range(iters):
                binary_array = [character for character in str(bin(i))][2:]
                mixed_mask = ["0"] * (len(fields) - len(binary_array)) + binary_array
                mask = [int(value) for value in mixed_mask]
                # Creates a binary mask the length of fields provided
                # for two fields, we need 4 iters, going over [(0, 0), (1, 0), (0, 1), (1, 1)]

                condition_value = [
                    {
                        "field": field if mask[index] else vector_field,
                        "filter_type": "exists",
                        "condition": "==" if mask[index] else "!=",
                        "condition_value": "",
                    }
                    for index, (field, vector_field) in enumerate(
                        zip(fields, vector_fields)
                    )
                ]
                filters.append(
                    {"filter_type": "or", "condition_value": condition_value}
                )

        else:  # Special Case when only 1 field is provided
            condition_value = [
                {
                    "field": fields[0],
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": "",
                },
                {
                    "field": vector_fields[0],
                    "filter_type": "exists",
                    "condition": "!=",
                    "condition_value": "",
                },
            ]
            filters.append({"filter_type": "or", "condition_value": condition_value})

        return filters

    @staticmethod
    def _encode_documents(
        documents, encoders: Dict[str, List[Base2Vec]], field_types: Dict[str, str]
    ):
        updated_documents = documents

        for dtype, vectorizers in encoders.items():
            for vectorizer in vectorizers:
                fields = [
                    field
                    for field, field_type in field_types.items()
                    if f"_{dtype}_" == field_type
                ]
                updated_documents = vectorizer.encode_documents(
                    documents=updated_documents, fields=fields
                )

        return updated_documents

    def _update_vector_metadata(self, metadata: List[str]) -> None:
        updated_metadata = self.datasets.metadata(self.dataset_id)["results"]

        if "_vector_" not in updated_metadata:
            updated_metadata["_vector_"] = {}

        now = (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()

        for vector_name in metadata:
            updated_metadata["_vector_"][vector_name] = now

        self.datasets.post_metadata(
            dataset_id=self.dataset_id, metadata=updated_metadata
        )

    def _insert_document_vectors(
        self,
        vector_fields: List[str],
        numeric_fields: List[str],
        show_progress_bar: bool,
    ):
        documents = self._get_all_documents(
            dataset_id=self.dataset_id,
            select_fields=vector_fields + numeric_fields,
            show_progress_bar=show_progress_bar,
        )
        schema = self._get_schema()
        if vector_fields:
            vectors = np.hstack(
                [
                    np.array(
                        [
                            self.get_field(vector_field, document)
                            if self.is_field(vector_field, document)
                            else [1e-7 for _ in range(schema[vector_field]["vector"])]
                            for document in documents
                        ]
                    )
                    for vector_field in vector_fields
                ]
            )
        else:
            vectors = np.array([])

        if numeric_fields:
            numbers = np.array(
                [
                    [
                        document[field] if field in document else 0
                        for field in numeric_fields
                    ]
                    for document in documents
                ]
            )
        else:
            numbers = np.array([])

        if numbers.size > 0 and vectors.size == 0:
            vectors = numbers

        elif numbers.size > 0 and vectors.size > 0:
            vectors = np.hstack([vectors, numbers])

        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler()
        vectors = scaler.fit_transform(vectors)
        vectors = np.nan_to_num(vectors)

        if vectors.shape[1] > 512:
            vectors = self._reduce(vectors)

        vectors = scaler.fit_transform(vectors)
        vectors = np.nan_to_num(vectors)

        document_vector_ = f"_dim{vectors.shape[1]}_feature_vector_"
        print(f"Concatenated field is called {document_vector_}")

        for index, (document, vector) in enumerate(zip(documents, vectors.tolist())):
            documents[index] = {"_id": document["_id"], document_vector_: vector}

        self._update_documents(
            dataset_id=self.dataset_id,
            documents=documents,
            show_progress_bar=show_progress_bar,
        )

        self._update_vector_metadata(metadata=[document_vector_])

    def run(
        self,
        dataset_id: str,
        fields: List[str],
        show_progress_bar: bool = True,
        detailed_schema: Optional[Dict[str, Any]] = None,
        filters: Optional[list] = None,
        **kwargs,
    ) -> None:
        if filters is None:
            filters = []
        self.dataset_id = dataset_id
        self.schema = self._get_schema()
        if detailed_schema is not None:
            self.detailed_schema = detailed_schema
        else:
            self.detailed_schema = self._get_detailed_schema()

        numeric_fields = self._get_numeric_fields()

        if fields:
            if "numeric" in fields:
                if len(fields) == 1:
                    field_types = {}

                else:
                    field_types = self._get_fields(fields)

            else:
                numeric_fields = [field for field in numeric_fields if field in fields]
                field_types = self._get_fields(fields)

        else:
            fields_list = list(self.detailed_schema)
            field_types = self._get_fields(fields_list)
            print(
                "No fields were given, vectorizing the following field(s): {}".format(
                    ", ".join(list(field_types))
                )
            )

        if field_types:
            self._validate_fields(list(field_types))

            self._init_encoders(list(field_types))

            vector_fields = self._get_vector_fields(list(field_types))
            print(
                "This operation will create the following vector_fields: {}".format(
                    str(vector_fields)
                )
            )

            filters += self._get_filters(
                fields=list(field_types), vector_fields=vector_fields
            )

            updating_args = dict(encoders=self.encoders, field_types=field_types)

            results = self.pull_update_push(
                dataset_id=self.dataset_id,
                update_function=self._encode_documents,
                select_fields=list(fields),
                filters=filters,
                show_progress_bar=show_progress_bar,
                updating_args=updating_args,
            )
            if results["failed_documents"]:
                print(Messages.INSERT_BAD)
                print(
                    "There were some errors vectorizing your unstructured data"
                )  # TODO: move messages in Messages class
                return results
        else:
            vector_fields = []

        new_schema = self._get_schema().keys()
        added_vectors = list(new_schema - self.schema)

        self._update_vector_metadata(metadata=added_vectors)

        if self.feature_vector:
            print(
                "Concatenating the following fields to form a feature vector: {}".format(
                    ", ".join(vector_fields + numeric_fields)
                )
            )
            self._insert_document_vectors(
                vector_fields=vector_fields,
                numeric_fields=numeric_fields,
                show_progress_bar=show_progress_bar,
            )

            new_schema = self._get_schema().keys()
            added_vectors = list(new_schema - self.schema)

        if added_vectors:
            print(Messages.INSERT_GOOD)
            print(
                "The following vector fields were added: " + ", ".join(added_vectors)
            )  # TODO: move messages in Messages class



"""
Visualisations for your clustering.
"""
import pandas as pd
import numpy as np
from relevanceai.constants.errors import (
    MissingClusterError,
    MissingPackageError,
    SetArgumentError,
)
from tqdm.auto import tqdm
from relevanceai.utils.decorators.analytics import track
from relevanceai.utils import largest_indices
from relevanceai.operations.cluster.cluster import ClusterOps
from typing import Any, Dict, List, Optional, Tuple, Union, Set, Callable
from relevanceai.operations.cluster.utils import ClusterUtils


class ClusterVizOps(ClusterOps, ClusterUtils):
    """
    Cluster Visualisations. May contain additional visualisation
    dependencies.
    """

    def __init__(
        self,
        credentials,
        dataset_id: str,
        vector_fields: List[str],
        alias: Optional[str] = None,
        **kwargs,
    ):
        self.vector_fields = vector_fields  # type: ignore
        self.alias = alias  # type: ignore
        self.dataset_id = dataset_id
        super().__init__(
            credentials=credentials,
            vector_fields=vector_fields,
            alias=alias,
            dataset_id=dataset_id,
            **kwargs,
        )

    @track
    def plot_basic_distributions(
        self,
        numeric_field: str,
        top_indices: int = 10,
        dataset_id: Optional[str] = None,
    ):
        """
        Plot the sentence length distributions across each cluster

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()

            cluster_ops = client.ClusterVizOps(
                dataset_id="sample_dataset",
                vector_fields=["sample_vector_"],
                alias="kmeans-5"
            )
            cluster_ops.plot_basic_distributions()

        Parameters
        ------------
        numeric_field: str
            The numeric field to plot
        top_indices: int
            The top indices in the plotting
        dataset_id: Optional[str]
            The dataset ID

        """
        try:
            import seaborn as sns
            import matplotlib.pyplot as plt
        except ModuleNotFoundError:
            raise MissingPackageError(package="seaborn")

        cluster_field = self._get_cluster_field_name()
        docs = self._get_all_documents(
            dataset_id=self.dataset_id if dataset_id is None else dataset_id,
            select_fields=[numeric_field, cluster_field],
        )
        df = pd.json_normalize(docs)
        top_comms = df[cluster_field].value_counts()
        for community in top_comms.index[:top_indices]:
            sample_comm_df = df[df[cluster_field] == community]
            sns.displot(sample_comm_df[numeric_field])
            # Get the average in the score too
            mean = sample_comm_df[numeric_field].mean()
            var = sample_comm_df[numeric_field].var()
            plt.title(
                community + str(f" - average: {round(mean, 2)}, var: {round(var, 2)}")
            )
            plt.show()

    @track
    def plot_distributions(
        self,
        numeric_field: str,
        measure_function: Callable = None,
        top_indices: int = 10,
        dataset_id: str = None,
        asc: bool = True,
        measurement_name: str = "measurement",
    ):
        """
        Plot the distributions across each cluster
        measure_function is run on each cluster and plots

        Example
        ----------
        .. code-block::

            from scipy.stats import skew
            ops.plot_distributions_measure(numeric_field, skew, dataset_id=dataset_id)

        Parameters
        -------------
        numeric_field: str
            The numeric field to plot the distribution by
        measure_function: callable
            What to measure the function
        top_indices: int
            The top indices
        dataset_id: str
            The dataset ID to use
        asc: bool
            If True, the distributions are plotted
        measurement_name: str
            The name of what should be plotted for the graphs

        """
        if measure_function is None:
            return self.plot_basic_distributions(
                numeric_field=numeric_field,
                top_indices=top_indices,
                dataset_id=dataset_id,
            )
        try:
            import seaborn as sns
            import matplotlib.pyplot as plt
        except ModuleNotFoundError:
            raise MissingPackageError(package="seaborn")

        cluster_field = self._get_cluster_field_name()

        # use the max and min to make the x axis the same
        if dataset_id is None:
            dataset_id = self.dataset_id
        numeric_field_facet = self.datasets.facets(
            dataset_id=dataset_id, fields=[numeric_field]
        )

        facet_result = numeric_field_facet["results"][numeric_field]

        docs = self._get_all_documents(
            dataset_id=dataset_id if dataset_id is None else dataset_id,
            select_fields=[numeric_field, cluster_field],
        )
        df = pd.json_normalize(docs)
        top_comms = df[cluster_field].value_counts()
        cluster_measurements = {}
        for community in tqdm(top_comms.index):
            sample_comm_df = df[df[cluster_field] == community]
            measure_output = measure_function(
                sample_comm_df[numeric_field].dropna().to_list()
            )
            cluster_measurements[community] = measure_output

        cluster_measurements = {
            k: v
            for k, v in sorted(
                cluster_measurements.items(), key=lambda item: item[1], reverse=asc
            )
        }

        for i, (community, measurement) in enumerate(cluster_measurements.items()):
            if i == top_indices:
                return
            sample_comm_df = df[df[cluster_field] == community]
            g = sns.displot(
                sample_comm_df[numeric_field],
            )
            g.set(xlim=(facet_result["min"], facet_result["max"]))
            plt.title(community + str(f" - {measurement_name}: {measurement}"))

    def plot_most_skewed(
        self,
        numeric_field: str,
        top_indices: int = 10,
        dataset_id: str = None,
        asc: bool = True,
    ):
        """
        Plot the most skewed numeric fields
        """
        from scipy.stats import skew

        return self.plot_distributions(
            numeric_field=numeric_field,
            measure_function=skew,
            top_indices=top_indices,
            dataset_id=dataset_id,
            asc=asc,
        )

    def _check_for_dataset_id(self):
        if not hasattr(self, "dataset_id"):
            raise SetArgumentError("dataset_id")

    def _get_cluster_field_name(self, alias: str = None):
        if alias is None:
            alias = self.alias
        if isinstance(self.vector_fields, list):
            set_cluster_field = f"_cluster_.{'.'.join(self.vector_fields)}.{alias}"
        elif isinstance(self.vector_fields, str):
            set_cluster_field = f"{self.cluster_field}.{self.vector_fields}.{alias}"
        elif self.vector_fields == None:
            raise ValueError("Vector field is not set.")
        else:
            raise ValueError("Can't detect cluster field.")
        return set_cluster_field

    def centroid_heatmap(
        self,
        metric: str = "cosine",
        vmin: float = 0,
        vmax: float = 1,
        print_n: int = 8,
        round_print_float: int = 2,
    ):
        """
        Heatmap visualisation of the closest clusters.
        Prints the ones ranked from top to bottom in terms of largest cosine similarity.
        """
        closest_clusters = self.closest(include_vector=True, verbose=False)
        import seaborn as sns
        from doc_utils import DocUtils
        from relevanceai.utils.cosine_similarity import cosine_similarity
        from sklearn.metrics import pairwise_distances

        shape = (len(closest_clusters["results"]), len(closest_clusters["results"]))
        all_vectors = []
        if self.vector_fields is not None:
            if len(self.vector_fields) == 1:  # type: ignore
                vector_field = self.vector_fields[0]
            else:
                raise NotImplementedError
        else:
            raise ValueError("Please set vector fields in the initialization.")

        for c, values in closest_clusters["results"].items():
            all_vectors.append(values["results"][0][vector_field])
        dist_out = 1 - pairwise_distances(all_vectors, metric="cosine")

        dist_df = pd.DataFrame(dist_out)
        heatmap_values = list(closest_clusters["results"].keys())
        dist_df.columns = heatmap_values
        dist_df.index = heatmap_values

        # ignore the initial set as they are the same indices
        ignore_initial = dist_out.shape[0]
        left, top = largest_indices(np.tril(dist_out), ignore_initial + print_n)

        left = left[ignore_initial:]
        top = top[ignore_initial:]

        print("Your closest centroids are:")
        for l, t in zip(left, top):
            print(
                f"{round(dist_out[l][t], round_print_float)} {heatmap_values[l]}, {heatmap_values[t]}"
            )

        return sns.heatmap(
            data=dist_df,
            vmin=vmin,
            vmax=vmax,
        ).set(title=f"{metric} plot")

    def show_closest(
        self,
        cluster_ids: Optional[List] = None,
        text_fields: Optional[List] = None,
        image_fields: Optional[List] = None,
    ):
        """
        Show the clusters with the closest.
        """
        from relevanceai import show_json

        if text_fields is None:
            text_fields = []
        if image_fields is None:
            image_fields = []
        new_closest = self.closest(cluster_ids=cluster_ids)
        closest_reformat = []
        for k, v in new_closest["results"].items():
            for r in v["results"]:
                closest_reformat.append({"cluster_id": k, **r})
        if cluster_ids is not None:
            text_fields += ["cluster_id"]
        return show_json(
            closest_reformat,
            text_fields=text_fields,
            image_fields=image_fields,
        )



"""
Base class for operations.
"""
from typing import Any, List, Union
from relevanceai.client.helpers import (
    Credentials,
    process_token,
)


class BaseOps:
    """
    Base class for operations
    """

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def init(self, *args, **kwargs):
        return self(*args, **kwargs)

    @classmethod
    def from_credentials(self, credentials: Credentials, *args, **kw):
        return self(credentials=credentials)

    @classmethod
    def from_token(self, token: str, *args, **kw):
        """
        If this is from a token, then we use this
        """
        credentials = process_token(token)
        return self(credentials=credentials, *args, **kw)

    @classmethod
    def from_client(self, client, *args, **kwargs):
        credentials = client.credentials
        return self(
            credentials=credentials,
            *args,
            **kwargs,
        )

    @classmethod
    def from_dataset(
        self, dataset: Any, alias: str, vector_fields: List[str], *args, **kwargs
    ):
        dataset_id = dataset.dataset_id
        credentials = dataset.credentials
        return self(
            credentials=credentials,
            dataset_id=dataset_id,
            alias=alias,
            vector_fields=vector_fields,
            vector_field=vector_fields[0],
            *args,
            **kwargs,
        )

    def _get_dataset_id(self, dataset: Union[str, Any]):
        from relevanceai.dataset import Dataset

        if isinstance(dataset, str):
            return dataset
        elif isinstance(dataset, Dataset):
            return dataset.dataset_id



from relevanceai.utils.decorators.analytics import *
from relevanceai.utils.decorators.thread import *
from relevanceai.utils.decorators.version import *
from relevanceai.utils.decorators.silence import *
from relevanceai.utils.decorators.log import *
from relevanceai.utils.decorators.vectors import *



"""
A simple CSV to dashboard workflow
"""
import pandas as pd
from typing import Optional, Dict, List
from relevanceai import ClusterOps


def df_to_dashboard(df: pd.DataFrame, text_cols: list):
    raise NotImplementedError


def text_csv_to_dashboard(
    csv_filename,
    text_cols: list,
    dataset_id: str,
    insertion_kwargs: Optional[dict] = None,
    text_encoder: Optional[dict] = None,
    vectorize_kwargs: Optional[dict] = None,
    cluster_model: Optional[str] = None,
    cluster_kwargs: Optional[dict] = None,
    token: Optional[str] = None,
):
    """
    This CSV takes in a file with a simple text column
    and provides a baseline insights/analytics dashboard.
    This can then be iterated on and improved.

    Parameters
    ------------
    csv_filename: str
        The CSV filename to insert
    text_columns: list
        The list of text columns
    """
    if cluster_kwargs is None:
        cluster_kwargs = {}
    if vectorize_kwargs is None:
        vectorize_kwargs = {}
    if insertion_kwargs is None:
        insertion_kwargs = {}

    from relevanceai import Client

    client = Client(token=token)

    print("Creating dataset...")
    ds = client.Dataset(dataset_id)

    print("Inserting CSV...")
    ds.insert_csv(csv_filename, **insertion_kwargs)

    print("Vectorize...")
    ds.vectorize(fields=text_cols, encoders={"text": text_encoder}, **vectorize_kwargs)

    print("Cluster...")
    cluster_ops: ClusterOps = ds.cluster(cluster_model, **cluster_kwargs)

    # Create suggested topics - summarize_closest alternative
    # Create a sample dashboard
    # Then summarize the closest ones
    print("Creating dashboard...")
    deployable_id = ds.deployables.create(dataset_id)
    print("Updating topics...")
    cluster_ops.summarize_closest(text_fields=text_cols, deployable_id=deployable_id)



import pytest

from sklearn.cluster import MiniBatchKMeans

from relevanceai import Client
from relevanceai import ClusterOps
from relevanceai.dataset import Dataset

from tests.globals.constants import NOT_IMPLEMENTED


class TestClusterOps:
    vector_field = "sample_1_vector_"

    def test_dataset_cluster_1(self, test_dataset: Dataset):
        alias = "cluster_test_1"
        test_dataset.cluster(
            model="kmeans",
            cluster_config=dict(n_clusters=3),
            alias=alias,
            vector_fields=[self.vector_field],
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def test_dataset_cluster_2(self, test_dataset: Dataset):
        alias = "cluster_test_2"
        test_dataset.cluster(
            model="kmeans",
            cluster_config=dict(n_clusters=10),
            alias=alias,
            vector_fields=[self.vector_field],
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def test_dataset_cluster_3(self, test_dataset: Dataset):
        alias = "cluster_test_3"
        test_dataset.cluster(
            model="optics",
            alias=alias,
            vector_fields=[self.vector_field],
            include_cluster_report=False,
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def test_dataset_cluster_4(self, test_dataset: Dataset):
        alias = "cluster_test_4"
        test_dataset.cluster(
            model=MiniBatchKMeans(n_clusters=10),
            alias=alias,
            vector_fields=[self.vector_field],
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def testClusterUtils(self, test_client: Client, test_dataset: Dataset):
        vector_field = "sample_1_vector_"
        alias = "kmeans-10"

        operator = test_client.ClusterOps(model="kmeans", n_clusters=10)
        operator(test_dataset, vector_fields=[vector_field])

        schema = test_dataset.schema
        assert f"_cluster_" in schema
        assert f"_cluster_.{vector_field}" in schema
        assert f"_cluster_.{vector_field}.{alias}" in schema

    @pytest.mark.skip(NOT_IMPLEMENTED)
    def test_list_closest(self, test_client: Client, test_dataset: Dataset):
        assert False

    @pytest.mark.skip(NOT_IMPLEMENTED)
    def test_list_furthest(self, test_client: Client, test_dataset: Dataset):
        assert False

    def test_merge(self, test_client: Client, test_dataset: Dataset):
        test_dataset.cluster(
            model="kmeans",
            n_clusters=3,
            alias="new_clustering",
            vector_fields=["sample_1_vector_"],
        )

        centroids = test_client.services.cluster.centroids.list(
            dataset_id=test_dataset.dataset_id,
            alias="new_clustering",
            vector_fields=["sample_1_vector_"],
        )["results"]
        assert len(centroids) == 3

        ops = ClusterOps.from_dataset(
            dataset=test_dataset,
            alias="new_clustering",
            vector_fields=["sample_1_vector_"],
        )

        ops.merge(cluster_labels=[0, 1], alias="new_clustering")

        centroids = test_client.services.cluster.centroids.list(
            dataset_id=test_dataset.dataset_id,
            alias="new_clustering",
            vector_fields=["sample_1_vector_"],
        )["results"]
        assert len(centroids) == 2
